#ifndef GRIDPOINTER_HH
#define GRIDPOINTER_HH

#include <iostream> 
#include <string>
#include <vector>
#include <assert.h>
#include <memory>
#include <map>

// Dune - include
#include <dune/common/parallel/mpihelper.hh>
#include <dune/grid/common/datahandleif.hh>
#include <dune/grid/common/gridenums.hh>
#include <dune/grid/common/intersectioniterator.hh>
#include <dune/alugrid/3d/iterator.hh>
#include <dune/grid/common/backuprestore.hh>

// My includes
#include "starcdreader.hh"
#include "freeContainer.hh"

template <class GridType,int restart> struct GridPointer;
template <class GridType>
struct GridPointer<GridType,0>
{
  //template< class GridImp, template < class > class IntersectionImp >
  //class Intersection;

  typedef typename GridType::LeafIntersection Intersection;
  typedef typename GridType::LeafIntersectionIterator IntersectionIterator;

  //typedef Dune::MPIHelper::MPICommunicator MPICommunicatorType;
  typedef GridType Grid;
  static const int dim = GridType::dimension;

  //--------------//
  // CONSTRUCTORS //
  //--------------//

  // Constructor from filename
  explicit GridPointer ( const std::string& filename, const Dune::MPIHelper& helper )
    :gridAutoPtr_(0),
     elParam_(),
     bndParam_(),
     bndId_(),
     blocks_(),
     emptyParam_(),
     nofElParam_( 0 ),
     haveBndParam_( false ),
     reader_(filename,helper)
  {
    
    //Dune::GridFactory<GridType> factory = StarCDReader::read(fileName, comm);
    //    StarCDReader<GridType> reader(filename,helper);
    initialize(reader_);
    
  };//end constructor from filename

  // Copy-Constructor
  GridPointer( const GridPointer &org )
    : gridAutoPtr_(org.gridAutoPtr_),
      elParam_(org.elParam_),
      vtxParam_(org.vtxParam_),
      bndParam_(org.bndParam_),
      bndId_(org.bndId_),
      emptyParam_( org.emptyParam_ ),
      nofElParam_(org.nofElParam_),
      nofVtxParam_(org.nofVtxParam_),
      haveBndParam_(org.haveBndParam_)
  {};// end of copy-constructor

  //-----------//
  // OPERATORS //
  //-----------//

  //dereference operator
  GridType& operator*(){
    return *gridAutoPtr_;
  };

  const GridType& operator*() const{
    return *gridAutoPtr_;
  };


  //pointer-to-member operator
  GridType* operator->(){
    return gridAutoPtr_.operator-> ();
  }
  
  const GridType* operator->() const{
    return gridAutoPtr_.operator-> ();
  }
  
  // number of parameters by codimension
  int nofParameters(int cdim) const {
    switch(cdim){
    case 0: return nofElParam_; break;
    case 1: return haveBndParam_; break;
    case GridType::dimension: return nofVtxParam_; break;
    }
    return 0;
  }
  
  //number of parameters by entity. It uses number of parameters by codimension
  template <class Entity>
  int nofParameters( const Entity & ) const{
    return nofParameters( static_cast< int >( Entity::codimension ) );
  }

  //number of parameters by intersection
  // template < class GridImp, template< class > class IntersectionImp > 
  // int nofParameters( const Intersection< GridImp,IntersectionImp> & intersection) const{
  //   return parameters( intersection ).size();
  // }
  int nofParameters( const Intersection & intersection) const{
    return parameters( intersection ).size();
  }
  
  //access to entities parameters
  template <class Entity>
  std::vector < double > &parameters(const Entity &entity) const{
    typedef typename GridType::LevelGridView GridView;
    GridView gridView = gridAutoPtr_->levelView( 0 );
    switch( (int)Entity::codimension ){
    case 0:
      if(nofElParam_ > 0){
	assert( (unsigned int)gridView.indexSet().index(entity) < elParam_.size() );
	return elParam_[ gridView.indexSet().index(entity) ];
      }
      break;
    case GridType::dimension:
      if(nofVtxParam_ > 0){
	assert( (unsigned int)gridView().indexSet().index(entity) < vtxParam_.size() );
	return vtxParam_[ gridView.indexSet().index(entity) ];
      }
      break;
    }
    return emptyParam_;
  }

  //set the number of parameters. to be used in datahandle class
  void setNofParams(int cdim, int nofP){
    switch(cdim){
    case 0: nofElParam_ = nofP; break;
    case 1: haveBndParam_ = nofP; break;
    case GridType::dimension: nofVtxParam_ = nofP; break;
    }
  }


  //acces to intersections parameters
  // template < class GridImp, template <class> class IntersectionImp>
  // const std::string & parameters( const Intersection< GridImp,IntersectionImp > & intersection ) const{
  //   if( !haveBndParam_ )
  //     return "";
  //   return bndParam_[ intersection.boundarySegmentIndex() ];
  // }
  const std::string & parameters( const Intersection & intersection ) const{
    if( !haveBndParam_ )
      return "";
    return bndParam_[ intersection.boundarySegmentIndex() ];
  }

  //GridPointer initialization
  void initialize(StarCDReader<GridType>& reader)
  {
    gridAutoPtr_ = std::auto_ptr< GridType > ( reader.factory_.createGrid(true,true,"testgrid2") );
    
    typedef typename GridType::LeafGridView GridView;
    GridView gridView = gridAutoPtr_->leafView();
    const typename GridView::IndexSet& indexSet = gridView.indexSet();

    //DEBUG
    // typedef typename GridView::template Codim< 0 >::template Partition<Dune::InteriorBorder_Partition>::Iterator LeafElementIterator;
    // typedef typename LeafElementIterator::Entity LeafEntity;
    // typedef typename GridType::template Codim<0>::EntityPointer LeafEntityPointer;
    // typedef typename GridView::IntersectionIterator LeafIntersectionIterator;
    // typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
    // LeafElementIterator eend = gridView.template end<0,Dune::InteriorBorder_Partition>();
    // for(LeafElementIterator eit = gridView.template begin<0,Dune::InteriorBorder_Partition>(); eit != eend; ++eit)
    //   {
    // 	LeafEntity & e = *eit;
    // 	std::cout << "element at " << e.geometry().center() << std::endl;
    // 	LeafIntersectionIterator iend = gridView.iend(e);
    // 	for(LeafIntersectionIterator iit = gridView.ibegin(e); iit != iend; ++iit)
    // 	  {
    // 	    const LeafIntersection & in = *iit;
    // 	    if(in.boundary())
    // 	      std::cout << "Intersection id " << in.boundaryId() << " at " << in.geometry().center() << std::endl;
    // 	  }
    // 	std::cout << "------------------" << std::endl;
    //   } 
    //DEBUG


    nofElParam_ = reader.template numberOfParameters< 0 >();
    nofVtxParam_ = reader.template numberOfParameters < dim >();
    haveBndParam_ = reader.haveBoundaryParameters();

    if ( nofElParam_ > 0 )
      elParam_.resize( indexSet.size(0) );
    if ( nofVtxParam_ > 0 )
      vtxParam_.resize( indexSet.size(dim) );
    bndId_.resize( indexSet.size(1) );
    if ( haveBndParam_ )
      bndParam_.resize( gridAutoPtr_->numBoundarySegments() );
    blocks_.resize(indexSet.size(0));

    const Dune::PartitionIteratorType partType = Dune::InteriorBorder_Partition;
    typedef typename GridView::template Codim< 0 >::template Partition< partType >::Iterator Iterator;
    const Iterator enditer = gridView.template end< 0, partType >();
    for( Iterator iter = gridView.template begin< 0, partType >(); iter != enditer; ++iter )
      {
	const typename Iterator::Entity &el = *iter;
	if( nofElParam_ > 0 ) {
	  //	  std::swap( elParam_[ indexSet.index(el) ], reader.parameter(el) );

	  //	  std::swap( elParam_[ indexSet.index(el) ], reader.elParams_[indexSet.index(el)] );
	  std::swap( elParam_[ indexSet.index(el) ], reader.elParams_[reader.factory_.insertionIndex(el)] );

	  if(elParam_[ indexSet.index(el) ].size()  != (size_t)nofElParam_)
	    std::cout << "Err: " << elParam_[ indexSet.index(el) ].size() << " " << (size_t)nofElParam_ << std::endl;
	  assert( elParam_[ indexSet.index(el) ].size()  == (size_t)nofElParam_ );
	}
	if( nofVtxParam_ > 0 ) {
	  for( unsigned int v = 0; v < el.subEntities( dim ); ++v )
	    {
	      typename GridView::IndexSet::IndexType index = indexSet.subIndex(el,v,dim);
	      if( vtxParam_[ index ].empty() )
		//		std::swap( vtxParam_[ index ],reader.parameter( *el.template subEntity< dim >(v) ) );
		std::swap( vtxParam_[ index ],reader.vtxParams_[ reader.factory_.insertionIndex( el.template subEntity<dim>(v)) ] );
	      assert( vtxParam_[ index ].size() == (size_t)nofVtxParam_ );
	    }
	}
	//INTERSECTION INITIALIZATION
	if( el.hasBoundaryIntersections() ){
	  //typedef typename GridView::LeafIntersectionIterator IntersectionIterator;
	  IntersectionIterator iend = gridView.iend( el );
	  for( IntersectionIterator iiter = gridView.ibegin( el ); iiter != iend; ++iiter){
	    const Intersection &inter = *iiter;
	    if( inter.boundary() && inter.type().isNone() && reader.factory_.wasInserted(inter) ){
	      const int k = indexSet.subIndex( el,inter.indexInInside(),1 );
	      bndId_[ k ] = reader.boundaryId( inter );
	      if( haveBndParam_ )
		//		bndParam_[ inter.boundarySegmentIndex() ] = reader.bndParams_[ inter.boundarySegmentIndex() ];
		bndParam_[ inter.boundarySegmentIndex() ] = reader.bndParams_[ reader.factory_.insertionIndex(inter) ];
	    }
	  }
	}
	std::swap( blocks_[ indexSet.index(el) ], reader.blocks_[ reader.factory_.insertionIndex(el) ] );
      }

    idBndToStringVector_=reader_.idBndToStringVector;

  };//end of initialize method

  //LoadBalance
  void loadBalance()
  {
    if( gridAutoPtr_->comm().size() == 1)
      return;
    int params = nofElParam_ + nofVtxParam_;
    if( haveBndParam_ )
      params += 1;
    if( gridAutoPtr_->comm().max( params ) > 0){
      ReaderDataHandle dh(*this);
      gridAutoPtr_->loadBalance( dh.interface() );
      gridAutoPtr_->communicate( dh.interface(),Dune::InteriorBorder_All_Interface,Dune::ForwardCommunication );
    }
    else{
      gridAutoPtr_->loadBalance();
    }
  }


  //************//
  //************//
  // DATAHANDLE //
  //************//
  //************//


  struct ReaderDataHandle : public Dune::CommDataHandleIF< ReaderDataHandle,double >
  {
    //NO INTERSECTION SUPPORT. intersection parameters are strings!!!!
    //CONSTRUCTOR
    ReaderDataHandle( GridPointer& gridPtr ) : 
      gridPtr_( gridPtr ),
      idSet_( gridPtr->localIdSet() )
    {
      //typedef typename GridType::LevelGridView GridView;
      typedef typename GridType::LeafGridView GridView;
      //GridView gridView = gridPtr_->levelView( 0 ); 
      GridView gridView = gridPtr_->leafView(); 
      const typename GridView::IndexSet &indexSet = gridView.indexSet();
      
      const Dune::PartitionIteratorType partType = Dune::InteriorBorder_Partition;
      typedef typename GridView::template Codim< 0 >::template Partition< partType >::Iterator Iterator;
      const Iterator enditer = gridView.template end< 0, partType >();
      for( Iterator iter = gridView.template begin< 0, partType >(); iter != enditer; ++iter )
	{
	  const typename Iterator::Entity &el = *iter;
	  if( gridPtr_.nofElParam_ > 0 ) 
	    std::swap( gridPtr_.elParam_[ indexSet.index(el) ], elData_[ idSet_.id(el) ] );
	  if( gridPtr_.nofVtxParam_ > 0){
	    for( unsigned int v = 0; v < el.subEntities( dim ); ++v){
	      typename GridView::IndexSet::IndexType index = indexSet.subIndex( el,v,dim );
	      if( !gridPtr_.vtxParam_[ index ].empty() )
		std::swap( gridPtr_.vtxParam_[ index ], vtxData_[ idSet_.subId( el,v,dim ) ]); 
	    }
	  }
	  std::swap( gridPtr_.blocks_[ indexSet.index(el) ], blockData_[ idSet_.id(el) ] );
	}
    };// end of DataHandle Constructor

    //DESTRUCTOR
    ~ReaderDataHandle()
    {
      //typedef typename GridType::LevelGridView GridView;
      typedef typename GridType::LeafGridView GridView;

      //GridView gridView = gridPtr_->levelView( 0 ); 
      GridView gridView = gridPtr_->leafView(); 
      const typename GridView::IndexSet &indexSet = gridView.indexSet();
      
      if ( gridPtr_.nofElParam_ > 0 )
	gridPtr_.elParam_.resize( indexSet.size(0) );
      if ( gridPtr_.nofVtxParam_ > 0 )
	gridPtr_.vtxParam_.resize( indexSet.size(dim) );
      gridPtr_.blocks_.resize( indexSet.size(0) );
      
      const Dune::PartitionIteratorType partType = Dune::All_Partition;
      typedef typename GridView::template Codim< 0 >::template Partition< partType >::Iterator Iterator;
      const Iterator enditer = gridView.template end< 0, partType >();
      for( Iterator iter = gridView.template begin< 0, partType >(); iter != enditer; ++iter )
	{
	  const typename Iterator::Entity &el = *iter;
	  if ( gridPtr_.nofElParam_ > 0 ) 
	    {
	      std::swap( gridPtr_.elParam_[ indexSet.index(el) ], elData_[ idSet_.id(el) ] );
	      assert( gridPtr_.elParam_[ indexSet.index(el) ].size() == (unsigned int)gridPtr_.nofElParam_ );
	    }
	  if( gridPtr_.nofVtxParam_ > 0)
	    {
	      for( unsigned int v = 0; v < el.subEntities( dim ); ++v){
		typename GridView::IndexSet::IndexType index = indexSet.subIndex( el,v,dim );
		if( !gridPtr_.vtxParam_[ index ].empty() )
		  std::swap( gridPtr_.vtxParam_[ index ], vtxData_[ idSet_.subId( el,v,dim ) ]);
		assert( gridPtr_.vtxParam_[ index ].size() == (unsigned int)gridPtr_.nofVtxParam_ );
	      }
	    }
	  std::swap( gridPtr_.blocks_[ indexSet.index(el) ], blockData_[ idSet_.id(el) ] );
	}
    };// end DataHandle Destructor

    //interface
    Dune::CommDataHandleIF< ReaderDataHandle,double > & interface(){
      return *this;
    }

    //contains
    bool contains( int dim, int codim ) const{
      return ( codim == 0 || codim ==dim );
    }
    
    //fixedsize?
    bool fixedsize( int dim, int codim ) const{
      return false;
    }
    
    //size
    template<class EntityType>
    size_t size(const EntityType& e) const{
      if(e.codimension==dim)
	return gridPtr_.nofParameters( (int)e.codimension );
      else
	return gridPtr_.nofParameters( (int)e.codimension ) + 1;
    }

    //GATHER
    template< class MessageBufferImp, class EntityType >
    void gather( MessageBufferImp& buff, const EntityType& e) const
    {
      const std::vector< double > &v = (e.codimension == 0)? elData_[ idSet_.id(e) ]:vtxData_[ idSet_.id(e) ];
      const size_t s = v.size();
      for( size_t i = 0; i < s; ++i)
	buff.write( v[ i ] );
      if(e.codimension==0)
	buff.write( (double)blockData_[ idSet_.id(e) ] );
      assert( s == (size_t)gridPtr_.nofParameters( e.codimension ));
    }

    //SCATTER
    template< class MessageBufferImp, class EntityType >
    void scatter( MessageBufferImp& buff, const EntityType& e, size_t n) const
    {
      std::vector< double > &v = (e.codimension == 0)? elData_[ idSet_.id(e) ]:vtxData_[ idSet_.id(e) ];
      if(e.codimension==dim){
	v.resize( n );
	gridPtr_.setNofParams( e.codimension,n );
	for(size_t i = 0; i < n; ++i)
	  buff.read( v[i] );
      }
      else
	{
	  v.resize( n-1 );
	  gridPtr_.setNofParams( e.codimension,n-1 );
	  for(size_t i = 0; i < n-1; ++i)
	    buff.read( v[i] );
	  double block;
	  buff.read(block);
	  blockData_[ idSet_.id(e) ] = (int)block;
	}
    }

  private:
    typedef typename GridType::LocalIdSet IdSet;
    GridPointer &gridPtr_;
    const IdSet &idSet_;
    mutable std::map< typename IdSet::IdType, std::vector< double > > elData_,vtxData_;
    mutable std::map< typename IdSet::IdType, int > blockData_;
  };// end of ReaderDataHandle struct

  //*************************//
  // END OF READERDATAHANDLE //
  //*************************//


  //EXPORT PARAMETERS TO VECTORS
  //for fields rows
  void exportElParamOneField( std::vector< double > & data,int indexParameter )
  {
    assert( indexParameter < nofElParam_);
    for(size_t i = 0; i < elParam_.size(); ++i)
      {
	data.push_back( elParam_[ i ][indexParameter] );
      }
  };
  void exportElParamAll( std::vector< std::vector< double > >& data )
  {
    data.resize(nofElParam_);
    for(size_t j = 0; j < nofElParam_; ++j)
      {
	for(size_t i = 0; i < elParam_.size(); ++i)
	  {
	    data[j].push_back( elParam_[ i ][ j ] );
	  }
      }
  };

  //DEBUG
  //for fields columns
  void exportElParamAllCols( std::vector< std::vector< double> >& data )
  {
    typedef typename GridType::LeafGridView GridView;
    GridView gridView = gridAutoPtr_->leafView(); 
    const typename GridView::IndexSet &indexSet = gridView.indexSet();
    
    if(nofElParam_ > 0){
      data.resize( indexSet.size( 0 ) );
      for(size_t i = 0; i < elParam_.size(); ++i)
	{
	  data[i].resize(nofElParam_);
	  for(size_t j = 0; j < nofElParam_; ++j)
	    {
	      //data[i].push_back( elParam_[ i ][ j ] );
	      data[i][j] = elParam_[i][j];
	    }
	}
      
      //probabilmente basta questo Verifica che data e elParam_ abbiano campi per colonna
      data = elParam_;
    }

  };

  //extract one field from data structure
  template < class M, class D>
  typename D::value_type extractField( D& data, int index, M& mapper)
  {
    typedef typename GridType::LeafGridView LeafGridView;
    typedef typename LeafGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator Iterator;
    typedef typename Iterator::Entity Element;

    typename D::value_type field( (size_t)mapper.size(),0.0 );

    LeafGridView leafGridView = gridAutoPtr_->leafView();
    Iterator end = leafGridView.template end<0,Dune::InteriorBorder_Partition>();
    for(Iterator it = leafGridView.template begin<0,Dune::InteriorBorder_Partition>(); it != end; ++it)
      {
	Element & e = *it;
	int indexi = mapper.map( e );
	field[indexi] = data[indexi][index];
      }
    return field;
  };

  //DEBUG
  
  // template <class T>
  // void freeContainer(T & t)
  // {
  //   T tmp;
  //   t.swap( tmp );
  // }
  

  //---------//
  // MEMBERS //
  //---------//
  // grid auto pointer
  mutable std::auto_ptr<GridType> gridAutoPtr_; // può essere modificato da un metodo const
  // reader
  StarCDReader<GridType> reader_;
  // element parameters
  std::vector< std::vector< double > > elParam_;
  std::vector< std::vector< double > > vtxParam_;
  std::vector< std::string > bndParam_;
  std::vector< int > bndId_;
  std::vector< double > emptyParam_;
  std::vector< int > blocks_;
  int nofElParam_;
  int nofVtxParam_;
  bool haveBndParam_;
  std::vector<std::string> idBndToStringVector_;

};//end GridPointer struct

#endif
