#ifndef DUNE_STARCD_READER_HH
#define DUNE_STARCD_READER_HH

#include <dune/common/exceptions.hh>
//#include <dune/common/geometrytype.hh>
#include <dune/geometry/type.hh>
//#include <dune/grid/common/gridfactory.hh>
#include <dune/alugrid/grid.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/grid/common/intersection.hh>
//#include "myboundarysegment.hh"
//#include "myParametrization.hh"
#include <iostream>
#include <fstream>
#include <limits>
#include "myPow.hh"

char **alloc_2d_char(int rows, int cols) 
{
  char *data = (char *)malloc(rows*cols*sizeof(char));
  char **array= (char **)malloc(rows*sizeof(char*));
  for (int i=0; i<rows; i++)
    array[i] = &(data[cols*i]);
  
  return array;
}

/** @ingroup StarCD
 *    \brief File reader for the Star-CD format.
 *
 *    Reads grid data described by the Star-CD format and fills 
 *    an empty grid with the data. Uses the grid creation methods
 *    described in 
 *    <a href="http://www.dune-project.org/doc/devel/ugreader.html">
 *    "How to Write a File Reader for UGGrid Objects"</a>.
 *    
 *    Two files \a fileName.vrt and \a fileName.cel have to be present. The file fileName.vrt contains the 
 *    coordinates of the nodes, each row having the format 
 *    \code 
 *    idx  x-coordinate  y-coordinate  z-coordinate 
 *    \endcode
 *    
 *    The file fileName.cel contains the data of the volume and possibly the boundary elements, each 
 *    row having the format
 *    \code 
 *    idx  node1  node2  ...  node8  material/boundaryId  flag1 flag2 
 *    \endcode
 *    The flags flag1 and flag2 appear to be always identical to 1 in case of a volume element 
 *    and to 4 in case of a boundary element. The element types simplex, pyramid, prism, and 
 *    cube are supported: 
 *    
 *    - For cubes, the indices node1 ... node8 are what you expect them to be. 
 *    - For simplices, node3 and node4 are identical, as well as node5 ... node8. 
 *    - For pyramids, node5 ... node8 are identical. 
 *    - For prisms, node3 and node4 are identical, as well as node7 and node8. 
 *
 *    This reader only supports three-dimensional grids.
 *    
 *    Currently no boundary element data is passed to \a grid. 
 */
template <class GridType>
class StarCDReader {
  
  typedef Dune::MPIHelper::MPICommunicator MPICommunicatorType;
  static const int dim = GridType::dimension;
  typedef typename GridType::ctype ctype;
  typedef typename GridType::template Codim< 0 >::Entity Element;
  typedef typename GridType::template Codim< dim >::Entity Vertex;
  template< class GridImp, template < class > class IntersectionImp >
  class Intersection;

public:

  //-------------//
  // CONSTRUCTOR //
  //-------------//
  
  StarCDReader ( const std::string& fileName, const Dune::MPIHelper& helper ,bool verbose = true)
  {

    // number of boundary types
    int nofBnd;
        
    // currently only dim = 3 is implemented
    if (dim != 3)
      DUNE_THROW(Dune::NotImplemented, 
		 "Reading Star-CD format is not implemented for dimension " << dim);
      
    if(helper.rank()==0){
      
      // PARAMETERS INFORMATION //**************************
      
      // set the name of the parameters file
      std::string parametersFileName = fileName + ".par";
      // set the parameters input stream
      std::ifstream parametersFile(parametersFileName.c_str());
      if (!parametersFile)
	DUNE_THROW(Dune::IOError, "Could not open " << parametersFileName);
      // reading the parameters file
      std::string parametersInfo;
      while(parametersFile >> parametersInfo){
	if(parametersInfo == "nofElParam"){
	  parametersFile >> nofElParameters_;
	  std::cout << "The number of element parameters is: " << nofElParameters_ << std::endl;
	}
	else if(parametersInfo == "nofVtxParam"){
	  parametersFile >> nofVtxParameters_;
	  std::cout << "The number of vertex parameters is: " << nofVtxParameters_ << std::endl;
	}
	else if(parametersInfo == "haveBndParam"){
	  parametersFile >> haveBndParameters_;
	  std::cout << "Has boundary parameters? " << haveBndParameters_ << std::endl;
	}
	else
	  std::cout << "Unknown item " << parametersInfo << " in " << parametersFileName << std::endl;
      }


      // VERTEX //*****************************************

      // set the name of the vertex file 
      std::string vertexFileName = fileName + ".vrt";
	
      // set the vertex input stream 
      std::ifstream vertexFile(vertexFileName.c_str());
      if (!vertexFile)
	DUNE_THROW(Dune::IOError, "Could not open " << vertexFileName);
	
      // read the vertices and parameters
      int dummyIdx;
      int numberOfVertices = 0;
      double buffParams = 0.0;
      std::vector < double > buffVtxParams;
      //auxiliary vector with vertex position: index+1 = insertion number.
      std::vector < Dune::FieldVector< ctype,dim > > vtxAuxVec;
      while (vertexFile >> dummyIdx) {
	numberOfVertices++;
	//reading vertex position         
	Dune::FieldVector<ctype,dim> position;
 	for (int k = 0; k < dim; k++)
	  vertexFile >> position[k];

	//fill the map (vertex coordinates,insertion number)
	vtxInsertOrder_.insert( std::make_pair(position,numberOfVertices-1) );
	//fill the auxiliary vector with vertex position
	vtxAuxVec.push_back(position);

	//reading parameters
	buffVtxParams.resize(0);
	for (int p = 0; p < nofVtxParameters_; p++){
	  vertexFile >> buffParams;
	  buffVtxParams.push_back(buffParams);
	}
	if(nofVtxParameters_ > 0)
	  vtxParams_.push_back(buffVtxParams);
	//insert vertex in the grid factory
	factory_.insertVertex(position);
      }
      if (verbose){
	std::cout << numberOfVertices << " vertices read." << std::endl;
	std::cout << vtxParams_.size() << " is the size of vtxParams_" << std::endl;
      }


      // ELEMENTS //********************************************

      // set the name of the element file 
      std::string elementsFileName = fileName + ".cel";
        
      // set the element input stream 
      std::ifstream elementsFile(elementsFileName.c_str());
      if (!elementsFile)
	DUNE_THROW(Dune::IOError, "Could not open " << elementsFileName);
	
      // read the elements 
      int numberOfElements = 0;
      int numberOfSimplices = 0;
      int numberOfPyramids = 0;
      int numberOfPrisms = 0;
      int numberOfCubes = 0;;
      int maxNumberOfVertices = (int)Pow(2, dim);
      int isVolume = 1;
      buffParams = 0.0;
      std::vector < double > buffElParams;
      while (elementsFile >> dummyIdx) {
	//element barycentre. It identifies univocally the element
	Dune::FieldVector< ctype,dim > elBarycentre(0);
	//reading element vertices and compute barycentre using vertices auxiliary vector
	std::vector<unsigned int> vertices(maxNumberOfVertices);
	for (int k = 0; k < maxNumberOfVertices; k++){
	  elementsFile >> vertices[k];
	  elBarycentre += vtxAuxVec[vertices[k]-1];
	}
	elBarycentre /= (double)maxNumberOfVertices;
	  
	//read two integer giving information about the block and the nature of the element
	int volumeOrSurface;
	int block;
	elementsFile >> block >> volumeOrSurface;
	blocks_.push_back(block);
          
	//if the element is a volume element then insert it in the factory
	numberOfElements++;
	//fill the element map (barycentre,insertion number)
	elInsertOrder_.insert( std::make_pair( elBarycentre, numberOfElements-1) );
	//identify the geometry of the element looking at its vertices
	if (vertices[2] == vertices[3]) { // simplex or prism
	  if (vertices[4] == vertices[5]) { // simplex
	    numberOfSimplices++;
	    std::vector<unsigned int> simplexVertices(4);
	    for (int k = 0; k < 3; k++)
	      simplexVertices[k] = vertices[k] - 1; 
	    simplexVertices[3] = vertices[4] - 1;
	    factory_.insertElement(Dune::GeometryType(Dune::GeometryType::simplex,dim), simplexVertices);
	  }
	  else { // prism
	    numberOfPrisms++;
	    std::vector<unsigned int> prismVertices(6);
	    for (int k = 0; k < 3; k++)
	      prismVertices[k] = vertices[k] - 1; 
	    for (int k = 3; k < 6; k++)
	      prismVertices[k] = vertices[k+1] - 1; 
	    factory_.insertElement(Dune::GeometryType(Dune::GeometryType::prism,dim), prismVertices);
	  }
	}
	else { // cube or pyramid 
	  if (vertices[4] == vertices[5]) { // pyramid
	    numberOfPyramids++;
	    std::vector<unsigned int> pyramidVertices(5);
	    for (int k = 0; k < 5; k++)
	      pyramidVertices[k] = vertices[k] - 1; 
	    factory_.insertElement(Dune::GeometryType(Dune::GeometryType::pyramid,dim), pyramidVertices);
	  }
	  else { // cube
	    numberOfCubes++;
	    std::vector<unsigned int> cubeVertices(8);
	    for (int k = 0; k < 8; k++)
	      cubeVertices[k] = vertices[k] - 1; 
	    std::swap(cubeVertices[2], cubeVertices[3]);
	    std::swap(cubeVertices[6], cubeVertices[7]);
	    factory_.insertElement(Dune::GeometryType(Dune::GeometryType::cube,dim), cubeVertices);
	    //factory_.insertElement(Dune::GeometryType(Dune::GeometryType::cube,dim), cubeVertices,Dune::shared_ptr<VirtualFunction<const FieldVector<ctype,dim>, FieldVector<ctype,dim> > >(new myParametrization<FieldVector<ctype,dim>, FieldVector<ctype,dim> >() ) );
	    
	  }
	}// end of if structure selecting element type (simplex/prism or cube/pyramid)
	
	//reading parameters
	buffElParams.resize(0);
	for (int p = 0; p < nofElParameters_; p++){
	  elementsFile >> buffParams;
	  buffElParams.push_back(buffParams);
	}

	if(nofElParameters_ > 0)
	  elParams_.push_back(buffElParams);
	
      }// end of while(elementsFile >> dummyIdx)
      if (verbose) {
	std::cout << numberOfElements << " elements read: " 
		  << numberOfSimplices << " simplices, " << numberOfPyramids << " pyramids, " 
		  << numberOfPrisms << " prisms, " << numberOfCubes << " cubes." << std::endl;
	std::cout << elParams_.size() << " is the size of elParams_" << std::endl;
      }
	
	
      // BOUNDARIES //*********************************************

      // set the name of the boundary file 
      std::string boundaryFileName = fileName + ".bnd";

      // set the boundary input stream 
      std::ifstream boundaryFile(boundaryFileName.c_str());
      if (!boundaryFile)
	std::cout << "NO BOUNDARY FILE" << std::endl; // NO EXCEPTION
	//	DUNE_THROW(Dune::IOError, "Could not open " << boundaryFileName);
      //CONSIDER NO BOUNDARY FILE	
      if(boundaryFile)
	{
	  // read the boundaries
	  int numberOfBoundaries = 0;
	  int numberOfQuadrilateral = 0;
	  int numberOfTriangles = 0;
	  int garbageInt;
	  int bnd5=0;
	  //string giving info about boundary condition
	  std::string bndString;
	  maxNumberOfVertices = (int)Pow(2, dim-1);
	  while (boundaryFile >> dummyIdx)
	    {
	      Dune::FieldVector< ctype,dim > bndBarycentre(0);	  
	      
	      //read boundary vertices
	      std::vector<unsigned int> vertices(maxNumberOfVertices); 
	      for (int k = 0; k < maxNumberOfVertices; k++){
		boundaryFile >> vertices[k];
		bndBarycentre += vtxAuxVec[ vertices[k]-1 ];
	      }
	      bndBarycentre /= maxNumberOfVertices;
	      
	      // read boundary identity
	      int boundaryId;
	      boundaryFile >> boundaryId;
	      if(boundaryId != 0){
		++numberOfBoundaries;
		bndInsertOrder_.insert( std::make_pair( bndBarycentre,numberOfBoundaries - 1 ) );
	      }
	      
	      //read garbage column???
	      boundaryFile >> garbageInt;// >> bndString;
	      
	      //read boundary parameters
	      if(haveBndParameters_ == 1){
		boundaryFile >> bndString;
		bndParams_.push_back(bndString);
	      }
	      
	      // if(boundaryId > idBndToString.size())
	      // 	idBndToString.resize(boundaryId,"");
	      idBndToStringMap[boundaryId] = bndString;
	      
	      if(vertices[2]==vertices[3])//Triangles
		{
		  std::vector<unsigned int> triVertices(maxNumberOfVertices - 1);
		  for(int k=0; k!=maxNumberOfVertices - 1; ++k)
		    triVertices[k]=vertices[k];
		  factory_.insertBoundary(Dune::GeometryType(Dune::GeometryType::simplex,dim-1),triVertices,boundaryId);
		}
	      else//Quadrilaterals
		{
		  std::vector<unsigned int> faceVertices(4);
		  for (int k = 0; k < 4; k++)
		    faceVertices[k] = vertices[k] - 1; 
		  std::swap(faceVertices[2], faceVertices[3]);
		  factory_.insertBoundary(Dune::GeometryType(Dune::GeometryType::cube,dim-1),faceVertices,boundaryId);
		}
	    } // end of while(boundaryFile >> dummyIdx)
	} // end of if(boundaryFile)
    } // end of if(helper.rank()==0)
    MPI_Barrier(helper.getCommunicator());

    nofBnd = idBndToStringMap.size();
    MPI_Bcast(&nofBnd, 1,MPI_INT,0,helper.getCommunicator());
    char idBndToStringArray[nofBnd*4];

    for(int i = 0; i < nofBnd; i++)
      {
	for(int j = 0; j < 4; j++)
	  {
	    idBndToStringArray[i*4+j] = idBndToStringMap[i+1][j];
	  }
      }
    idBndToStringMap.clear();
    
    MPI_Barrier(helper.getCommunicator());
    
    MPI_Bcast(idBndToStringArray, nofBnd*4,MPI_CHAR,0,helper.getCommunicator());
    
    std::string stringa4 = "zzzz";
    idBndToStringVector.resize(nofBnd,stringa4);
    for(int i = 0; i < nofBnd; ++i )
      {
	for(int j = 0; j < 4; j++)
	  {
	    idBndToStringVector[i].at(j) = idBndToStringArray[i*4+j];
	  }
      }

  } // end of StarCDReader constructor

  //---------//
  // METHODS //
  //---------//

  //NUMBER OF PARAMETERS WITH CODIMENSION
  template<int codim>
  int numberOfParameters(){
    if(codim==dim)
      return nofVtxParameters_;
    return nofElParameters_;
  };

  bool haveBoundaryParameters(){
    if(haveBndParameters_ > 0)
      return true;
    return false;
  }

  //EXTRACT PARAMETERS FROM ELEMENT

  std::vector < double >& parameter (const Element& element){
    const typename Element::Geometry &geo = element.geometry();
    DomainType coord( geo.corner( 0 ) );
    for( int i = 1; i < geo.corners(); ++i )
      coord += geo.corner( i );
    coord /= double( geo.corners() );
    
    InsertOrderIterator it = elInsertOrder_.find( coord );
    size_t counter = elInsertOrder_.count(coord);
    if( it != elInsertOrder_.end() )
      return elParams_[ it->second ];
    //    assert(0);
    return emptyParams_;
  }

  //EXTRACT PARAMETERS FROM VERTEX
  std::vector< double >& parameter(const Vertex& vertex){
    const typename Vertex::Geometry &geo = vertex.geometry();
    DomainType coord(geo.corner( 0 ));
    
    InsertOrderIterator it = vtxInsertOrder_.find( coord );
    if( it != vtxInsertOrder_.end() )
      return vtxParams_[ it->second ];
    return emptyParams_;
  }

  //EXTRACT PARAMETERS FROM INTERSECTION
  //template <class GridImp, template < class > class IntersectionImp>
  template <class I>
  const std::string & boundaryParameter(const I & intersection ) const {
    const typename I::Geometry &geo = intersection.geometry();
    DomainType coord( geo.corner( 0 ) );
    for(int i = 1; i < geo.corners(); ++i)
      coord += geo.corner( i );
    coord /= double( geo.corners() );
    
    InsertOrderIterator it = bndInsertOrder_.find( coord );
    if( it != bndInsertOrder_.end() )
      return bndParams_[ it->second ];
    return emptyString_;
  }

  //EXTRACT BOUNDARY ID FROM INTERSECTION
  template <class Intersection>
  int boundaryId(const Intersection& intersection) const
  {
    return intersection.boundaryId();
  }

  //--------------------------//
  // OTHER CLASSES AND STRUCT //
  //--------------------------//
  
  typedef Dune::FieldVector<typename GridType::ctype,GridType::dimensionworld> DomainType;
  struct Compare 
     { 
       bool operator() ( const DomainType &a, const DomainType &b ) const
       {
         // returns true, if a < b; c[i] < -eps;
         const DomainType c = a - b;
         const double eps = 1e-6;//std::numeric_limits<double>::epsilon();//1e-12;
	 
         for( int i = 0; i < DomainType::dimension; ++i )
  	   {
  	     if( c[ i ] <= -eps )
  	       return true;
  	     if( c[ i ] >= eps )
  	       return false;
  	   }
         return false;
       }
  };
  // struct Compare 
  //    { 
  //      bool operator() ( const DomainType &a, const DomainType &b ) const
  //      {
  //        // returns true, if a < b; c[i] < -eps;
  //        const DomainType c = a - b;
  //        const double eps = 1e-8;
  // 	 bool equal[DomainType::dimension];
  // 	 bool vecTrue[DomainType::dimension];
	 
  //        for( int i = 0; i < DomainType::dimension; ++i )
  // 	   {
  // 	     vecTrue[i] = true;
  // 	     if( c[ i ] <= eps && c[i]>=-eps)
  // 	       equal[i] = true;
  // 	     if( c[ i ] >= eps && c[i]<=-eps)
  // 	       equal[i] = false;
  // 	   }
  // 	 if( memcmp(equal,vecTrue,sizeof(equal)) == 0 )
  // 	   return true;
  //        return false;
  //      }
  // };
  typedef std::map< DomainType, size_t, Compare > InsertOrderMap;
  typedef typename InsertOrderMap::const_iterator InsertOrderIterator;

  //---------//
  // MEMBERS //
  //---------//
  Dune::GridFactory<GridType> factory_;
  int nofElParameters_;
  int nofVtxParameters_;
  bool haveBndParameters_;
  InsertOrderMap elInsertOrder_;
  InsertOrderMap vtxInsertOrder_;
  InsertOrderMap bndInsertOrder_;
  std::vector < std::vector < double > > elParams_;
  std::vector < std::vector < double > > vtxParams_;
  std::vector < int > bndId_;
  std::vector < int > blocks_;
  std::vector < std::string > bndParams_;
  std::vector < double > emptyParams_;
  std::string  emptyString_;
  std::map<int,std::string> idBndToStringMap;
  std::vector<std::string> idBndToStringVector;

};



#endif
